/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.repository.mapping;

import java.io.*;
import java.math.BigDecimal;
import java.sql.*;
import java.util.Arrays;
import java.util.Iterator;

import org.xml.sax.SAXException;
import org.xquark.repository.RepositoryException;
import org.xquark.repository.dbms.*;
import org.xquark.repository.metadata.RepositoryConstants;
import org.xquark.schema.Declaration;
import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.*;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLErrorHandler;

/** This class describes features of the database corresponding type
 * for a datum to map, basing on the XML schema type.
 *
 * <p>Performs checks between XML type and SQL type.</p>
 * 
 * <p>The type info build is the natural mapping between schema types
 * and JDBC. It is not built from the database metadata but will be
 * checked against it.</p>
 */
public class MappingTypeInfo extends JavaTypeInfo
{
    private static final String RCSRevision = "$Revision: 1.5 $";
    private static final String RCSName     = "$Name:  $";
    
    // mapping between schema type and JDBC base types
    public static int TYPE_MAPPINGS[] = new int[33]; // (index 0 unused)
    
    /** Conversion authorized by setObject */
    public static int AUTHORIZED_CONVERSIONS[][] = new int[9][]; // first index is JAVA object class

    static
    {
        // mapping between schema type and JDBC base types
        TYPE_MAPPINGS[PrimitiveType.STRING] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.BOOLEAN] = Types.DECIMAL;
        TYPE_MAPPINGS[PrimitiveType.DECIMAL] = Types.DECIMAL;
        TYPE_MAPPINGS[PrimitiveType.FLOAT] = Types.FLOAT;
        TYPE_MAPPINGS[PrimitiveType.DOUBLE] = Types.DOUBLE;
        TYPE_MAPPINGS[PrimitiveType.QNAME] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.NOTATION] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.DURATION] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.DATE_TIME] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.TIME] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.DATE] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.GYEAR_MONTH] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.GYEAR] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.GMONTH_DAY] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.GDAY] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.GMONTH] = Types.TIMESTAMP;
        TYPE_MAPPINGS[PrimitiveType.HEX_BINARY] = Types.VARBINARY;
        TYPE_MAPPINGS[PrimitiveType.BASE64_BINARY] = Types.VARBINARY;
        TYPE_MAPPINGS[PrimitiveType.ANY_URI] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.UNION] = Types.VARCHAR; // For generation, a VARCHAR column, may accept almost anything...
        TYPE_MAPPINGS[PrimitiveType.LIST] = Types.VARCHAR;
        TYPE_MAPPINGS[PrimitiveType.ANYSIMPLETYPE] = Types.VARCHAR;
        
        // Conversion authorized by setObject ("getting Started" of JDBC user guide)
        AUTHORIZED_CONVERSIONS[JAVA_STRING] = new int[]
        {
            Types.TINYINT, Types.SMALLINT, Types.INTEGER, Types.BIGINT,
            Types.REAL, Types.FLOAT, Types.DOUBLE, Types.DECIMAL, Types.NUMERIC,
            Types.BIT, Types.CHAR, Types.VARCHAR, Types.LONGVARCHAR,
            Types.BINARY, Types.VARBINARY, Types.LONGVARBINARY, Types.DATE,
            Types.TIME, Types.TIMESTAMP
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_STRING]); // for binary search
        // Note: this sort could be avoided but is safer if arrays are modified later
        // further more is precompiled-data
        
        AUTHORIZED_CONVERSIONS[JAVA_BIG_DECIMAL] =
        AUTHORIZED_CONVERSIONS[JAVA_BOOLEAN] =
        AUTHORIZED_CONVERSIONS[JAVA_FLOAT] =
        AUTHORIZED_CONVERSIONS[JAVA_DOUBLE] = new int[]
        {
            Types.TINYINT, Types.SMALLINT,
            Types.INTEGER, Types.BIGINT, Types.REAL, Types.FLOAT, Types.DOUBLE,
            Types.DECIMAL, Types.NUMERIC, Types.BIT, Types.CHAR, Types.VARCHAR,
            Types.LONGVARCHAR
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_DOUBLE]); // for binary search
        
        AUTHORIZED_CONVERSIONS[JAVA_BYTE_ARRAY] = new int[]
        {
            Types.BINARY,
            Types.VARBINARY, Types.LONGVARBINARY
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_BYTE_ARRAY]); // for binary search
        
        AUTHORIZED_CONVERSIONS[JAVA_TIMESTAMP] = new int[]
        {
            Types.CHAR,
            Types.VARCHAR, Types.LONGVARCHAR, Types.DATE, Types.TIME,
            Types.TIMESTAMP
        };
        Arrays.sort(AUTHORIZED_CONVERSIONS[JAVA_TIMESTAMP]); // for binary search
    }

    /* setString() and getString() can be used */
    private boolean isJDBCStringIOAllowed = true; // for default mapping
    
    /* setCharacterStream() and set() must be used */
    private boolean useStreams = false; // for default mapping
    
    /**
     * No conversion to performs since the schema output fits direcly the
     * column JDBC type.
     */
    private static final short NONE = 0;
    /** JDBC performed */
    private static final short IMPLICIT = 1;
    /**
     * Performed by mapper specific code. See
     * {@link #storeConvertedToSQL(PreparedStatement, int, Object)}.
     */
    private static final short AUTOMATIC = 2;
    
    /** User-defined (NYI) */
    private static final short EXPLICIT = 3;
    
    /** Forbidden */
    private static final short IMPOSSIBLE = 4;
    
    /** Flag used to determine if a  conversion is needed between the schema
     * output and JDBC
     */
    private short conversion = IMPOSSIBLE; // for default mapping
    
    private ColumnMetaData cmeta = null;
    private boolean targetIsString;
    /* use string delimitor if target is String */
    private boolean useStringDelimitor = false;
    
    //////////////////////////////////////////////////////////////////////////
    //                 CONSTRUCTORS & ACCESSORS
    //////////////////////////////////////////////////////////////////////////
    /** Constructor used for user generators
     */
    public MappingTypeInfo(String XMLType)
    throws SAXException
    {
        super(XMLType);
    }
    
    /* Used for value generators */
    public MappingTypeInfo(SimpleType XMLType)
    {
        super(XMLType);
    }
    
    public MappingTypeInfo(String XMLType, boolean useStringDelimitor)
    throws SAXException
    {
        super(XMLType);
        this.useStringDelimitor = useStringDelimitor;
    }
    
    /* Used for value generators */
    public MappingTypeInfo(SimpleType XMLType, boolean useStringDelimitor)
    {
        super(XMLType);
        this.useStringDelimitor = useStringDelimitor;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                      PUBLIC METHODS
    //////////////////////////////////////////////////////////////////////////
    void checkTargetType(ColumnMetaData cmeta)
    throws SAXException
    {
        checkTargetType(cmeta, null);
    }
    
    void checkTargetType(ColumnMetaData cmeta, XMLErrorHandler messageHandler)
    throws SAXException
    {
        // keep columnMetadata
        this.cmeta = cmeta;
        PrimitiveType pType = sType.getPrimitive();
        
        if (cmeta.isLongType())
            useStreams = true;
        
        if (pType.getType() == PrimitiveType.UNION)   // UNION
        {
            Iterator it = sType.getMemberTypes().iterator();
            PrimitiveType wType = null;
            
            while (it.hasNext())
            {
                conversion = IMPOSSIBLE;    // other flags do not use default value
                wType = ((SimpleType)it.next()).getPrimitive();
                checkTargetType(wType, JAVA_TYPE_BINDINGS[wType.getType()],
                                    messageHandler);
                // TO IMPROVE : do global flags have to be consistent for every type of Union ? What if not ? The last are kept ?
            }
        }
        else
            checkTargetType(pType, javaType, messageHandler);
    }
    
    /**
     * true if target is a SQL String Type or the schema convert output is a
     * Java String. In other words, generator returns a value that is of type
     * String because the actual value is a String or because the normalized or
     * canonical value is used.
     */ 
    public boolean useJDBCStringMethods()
    {
        return isJDBCStringIOAllowed;
    }
    
	/**
	 * true if target is a SQL String Type and the schema convert output is not
	 * a Java String.
	 */ 
    public boolean storeCanonicalString()
    {
        return isJDBCStringIOAllowed && (javaType != JAVA_STRING);
    }
    
    /** Schema conversion is supposed to be done
     * @param useEmptyStringAlias
     */
    public void setParameter(
                                PreparedStatement pStmt,
                                int i,
                                Object value,
                                AbstractConnection connection,
                                boolean useEmptyStringAlias,
                                XMLErrorHandler messageHandler
                                )
    throws SQLException, XMLDBCException
    {
        if (value == null)   // for Oracle : because setObjet(i, null) leads to an exception
            pStmt.setNull(i, cmeta.getDataType());
        else
        {
            if (conversion == AUTOMATIC)
                storeConvertedToSQL(pStmt, i , adaptSchemaObjectToJDBC(value));
            else if (useJDBCStringMethods())
            {
                String str = (String)value;

                if (targetIsString)
                {
                    if (useStringDelimitor)
                        str += RepositoryConstants.STRING_DELIMITOR;
                
                    if (useEmptyStringAlias && (str.length() == 0))
                        pStmt.setString(i, RepositoryConstants.STRING_DELIMITOR);
                    else if (str.length() > cmeta.getColumnSize())
                        throw new RepositoryException(RepositoryException.DATA_LOSS, "Value is too large for column.");
                    else
                    {
                        if (useStreams)
                            pStmt.setCharacterStream(i, new StringReader(str), str.length());
                        else
                            pStmt.setString(i, str);
                    }
                }
                else
                    pStmt.setString(i, str);
            }
            else    // object other than string (converted)
            {
                checkDataLoss(value, messageHandler);
                value = adaptSchemaObjectToJDBC(value);
                if (useStreams && (javaType == SCHEMA_BYTE_ARRAY))
                {
                    byte[] bytes = (byte[])value;
                    pStmt.setBinaryStream(i, new ByteArrayInputStream(bytes), bytes.length);
                }
                else
                    connection.setObject(pStmt, i, value, javaType, cmeta);
            }
        }
    }

    private Object adaptSchemaObjectToJDBC(Object value) {
        
        switch (javaType) {
            case SCHEMA_DURATION:
			case SCHEMA_URI:
			case SCHEMA_QNAME:
                value = value.toString();
				break;
				
			case SCHEMA_DATETIME:
				value = new Timestamp(((DateTime)value).getTime());
				break;
			
			case SCHEMA_BYTE_ARRAY:
				value = ((ByteArray)value).getData();
				break;
				
			case JAVA_LIST:
                value = sType.toXMLString(value);
				break;

            default :
                break;
        }
        return value;
    }

    /* Should be set in the RDBMS portability layer in AbstractStatement (to ADD) */
    public String getParameter(ResultSet rs, int i, Declaration decl, ValidationContextProvider nsContext)
    throws SQLException, XMLDBCException
    {
        String ret = null;
        SimpleType type = null;
        
        if (decl != null)   // for Text nodes or generic mapping
            type = decl.getType().getValueType();
        
        if (type == null)   // if no type, take the pseudo type from the table definition
            type = sType;
        
        if (useJDBCStringMethods() && (conversion != AUTOMATIC))
        {
            if (useStreams) // use getCharacterStream because Oracle does not tolerate getString for CLOBs
            {
                StringBuffer sb = new StringBuffer();
                Reader reader = rs.getCharacterStream(i);
                char[] buffer = new char[256];
                int len = 0;
                try
                {
                    while ((len = reader.read(buffer)) > 0)
                        sb.append(buffer, 0, len);
                }
                catch (IOException e)
                {
                    throw new SQLException("IOException while using read() on a Reader object obtained by getCharacterStream().");
                }
                ret = sb.toString();
            }
            else
            {
                ret = rs.getString(i);
                if (targetIsString)
                {
                    if (useStringDelimitor)
                        ret = ret.substring(0, ret.length() - RepositoryConstants.STRING_DELIMITOR_LENGTH);
                    else if ((ret != null) && ret.equals(RepositoryConstants.STRING_DELIMITOR))
                        ret = "";
                }
                if (!storeCanonicalString()) // data is not stored converted : canonize now
                    ret = type.toXMLString(ret, nsContext);
            }
        }
        else
        {
            Object o;
            if (conversion == AUTOMATIC)
                o = convertToXML(rs, i);
            else if (useStreams)    // BLOB or LONGVARBINARY
            {
                ByteArrayOutputStream out = new ByteArrayOutputStream(256);
                InputStream in = rs.getBinaryStream(i);
                byte[] buffer = new byte[256];
                int len = 0;
                try
                {
                    while ((len = in.read(buffer)) > 0)
                        out.write(buffer, 0, len);
                }
                catch (IOException e)
                {
                    throw new SQLException("IOException while using read() on a Reader object obtained by getCharacterStream().");
                }
                o = out.toByteArray();
            }
            else
                o = rs.getObject(i);    // because first index starts at 0 and OID belongs to columns
            
            if (o != null)
                ret = type.toXMLString(o, nsContext);
        }

        return ret;
    }
    
    /* generated object type must be insensitive to a setObject-getObject U-turn */
    private Object convertToXML(ResultSet rs, int i)
    throws SQLException, XMLDBCException
    {
        switch (cmeta.getDataType()) {
            case Types.VARBINARY :

                try {
                    return new String(rs.getBytes(i), "UTF-8");
                }
                catch (UnsupportedEncodingException e) {
                    throw new RuntimeException("UTF-8 encoding is not supported by the current JRE. Mapping from string to binary is impossible");
                }
            case Types.BIGINT :
            case Types.DECIMAL :
            case Types.DOUBLE :
            case Types.FLOAT :
            case Types.INTEGER :
            case Types.NUMERIC :
            case Types.REAL :
            case Types.SMALLINT :
            case Types.TINYINT :

                Timestamp ts = null;
                int n = rs.getInt(i);
                switch (sType.getPrimitive().getType()) {
                    case PrimitiveType.GYEAR :
                        ts = new Timestamp(n, 0, 0, 0, 0, 0, 0);
                        break;
                    case PrimitiveType.GMONTH :
                        ts = new Timestamp(0, n, 0, 0, 0, 0, 0);
                        break;
                    case PrimitiveType.GDAY :
                        ts = new Timestamp(0, 0, n, 0, 0, 0, 0);
                        break;
                    default :
                        }
                return ts;

            default :
                throw new RepositoryException(
                    RepositoryException.CONSISTENCY_ERROR,
                    "This automatic conversion is not currently supported (javaType: "
                        + javaType
                        + ", JDBCtype: "
                        + cmeta.getDataType()
                        + ')');
        }
    }
    
    public String getTypedNullValue() throws RepositoryException
    {
        switch (cmeta.getDataType())
        {
            case Types.BIT: case Types.TINYINT: case Types.SMALLINT:
            case Types.INTEGER: case Types.BIGINT: case Types.FLOAT:
            case Types.REAL: case Types.DOUBLE: case Types.NUMERIC:
            case Types.DECIMAL:
                return "TO_NUMBER(null)";
            case Types.CHAR: case Types.VARCHAR: case Types.LONGVARCHAR:
            case Types.CLOB:
                return "''";
            case Types.BINARY: case Types.VARBINARY: case Types.LONGVARBINARY:
            case Types.BLOB:
                return "0";
            case Types.DATE: case Types.TIME: case Types.TIMESTAMP:
                return "TO_DATE(null)";
            default:
                throw new RepositoryException(RepositoryException.NOT_IMPLEMENTED,
                "Such query cannot be handled because of the column " + cmeta.getDataType()
                + " SQL type");
        }
    }
        
    public String getTypeCreationString(AbstractConnection connection)
    {
        PrimitiveType pType = sType.getPrimitive();
        int JDBCType = TYPE_MAPPINGS[pType.getType()];
        
        int optimumJDBCType = optimize(
                JDBCType,
                connection.getTypeMap(),
                pType.getMaxLength(),
                pType.getTotalDigits(),
                pType.getFractionDigits(),
                (Number)pType.getMinValue(),
                (Number)pType.getMaxValue()
                );
        
        return ColumnSpec.getTypeCreationString(
                optimumJDBCType,
                connection.getTypeMap(),
                pType.getMaxLength(),
                pType.getTotalDigits(),
                pType.getFractionDigits()
                );
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                      PRIVATE UTILITIES
    //////////////////////////////////////////////////////////////////////////
    private void checkTargetType(PrimitiveType pType, int javaType,
                        XMLErrorHandler messageHandler)
    throws SAXException
    {
        int targetJDBCType = cmeta.getDataType();
        
        targetIsString = (targetJDBCType == Types.CHAR)
                        || (targetJDBCType == Types.VARCHAR)
                        || (targetJDBCType == Types.LONGVARCHAR);
        
        
		// set the isJDBCStringIOAllowed flag
		isJDBCStringIOAllowed = targetIsString || (javaType == JAVA_STRING);
        
        // First verify that setObject is authorized (redundant ?)
        if (javaType < JAVA_LIST
        		&& (Arrays.binarySearch(AUTHORIZED_CONVERSIONS[javaType], (int)targetJDBCType) < 0)    // Warning getDataType returns a short
                && !isJDBCStringIOAllowed)
            throw new SAXException(sType + " XML data type cannot be mapped on "
            + cmeta + " with SQL type " + cmeta.getTypeName() + ".");
        
        // Verify that the XML and SQL type fit
        
        // 1. Classify the XML type in a basic (JDBC) type within (VARCHAR,
        // VARBINARY, DECIMAL, FLOAT, DOUBLE, TIMESTAMP)
       int abstractBaseJDBCType = TYPE_MAPPINGS[pType.getType()];

       // conversion type flag setting
       if (abstractBaseJDBCType == targetJDBCType)
           conversion = NONE;
       else
       {
           switch (targetJDBCType)
           {
               case Types.VARCHAR:
                   conversion = IMPLICIT;
                   break;
                   
               case Types.VARBINARY:
                   if (abstractBaseJDBCType == Types.VARCHAR)
                       conversion = AUTOMATIC;
                   break;
                   
               case Types.DECIMAL:
                   if ((abstractBaseJDBCType == Types.REAL)
                    || (abstractBaseJDBCType == Types.FLOAT)
                    || (abstractBaseJDBCType == Types.DOUBLE))
                       conversion = IMPLICIT;
                   else if ((pType.getType() == PrimitiveType.GYEAR)
                    || (pType.getType() == PrimitiveType.GMONTH)
                    || (pType.getType() == PrimitiveType.GDAY))
                       conversion = AUTOMATIC;
                   break;
                   
               case Types.REAL:
                   if (abstractBaseJDBCType == Types.DECIMAL)
                       conversion = IMPLICIT;
                   break;
                   
               case Types.DOUBLE:
                   if ((abstractBaseJDBCType == Types.DECIMAL)
                    || (abstractBaseJDBCType == Types.FLOAT)
                    || (abstractBaseJDBCType == Types.REAL))
                       conversion = IMPLICIT;
                   break;
                   
               default:                   
           }
           
           if (conversion == IMPOSSIBLE)
               throw new SAXException(sType + " XML data type cannot be mapped on "
               + cmeta + " with SQL type " + cmeta.getTypeName() + ".");
       }
     
       // pseudo facet setting (could be performed only if the target type requires it)
       int  maxLen = pType.getMaxLength(),
            precision = pType.getTotalDigits(),
            scale = pType.getFractionDigits(); // if -1, no need to check
       
       switch (pType.getType())
       {
           case PrimitiveType.BOOLEAN:
               maxLen = 5;
               break;
           case PrimitiveType.DECIMAL:
               // maxlen calculation for writing in VARCHAR SQL type
               if (precision > 0) // use absolute limits if any
                   maxLen = precision + (scale > 0?2:1);   // sign and dot
               else if (scale >= 0) // if no scale and no precision no prediction on string size can be performed
               {
                   if ((pType.getMinValue() != null) && (pType.getMaxValue() != null))    // bounded
                   {
                       double maxAbs = Math.max(
                                        Math.abs(((Number)pType.getMinValue()).doubleValue()),
                                        Math.abs(((Number)pType.getMaxValue()).doubleValue())
                                        );
                       maxLen = (int)Math.ceil(Math.log(maxAbs)/Math.log(10d));
                       if (scale > 0)
                           maxLen += scale + 2;
                       else
                           maxLen += scale + 1;
                   }
               }
               break;
               
           case PrimitiveType.DATE_TIME:
               maxLen = 19; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.TIME:
               maxLen = 8; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.DATE:
               maxLen = 10; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GYEAR_MONTH:
               maxLen = 7; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GYEAR:
               maxLen = 4; // this is the minimum value for a date string using schema syntax
               precision = 4;
               scale = 0;
               break;
           case PrimitiveType.GMONTH_DAY:
               maxLen = 7; // this is the minimum value for a date string using schema syntax
               break;
           case PrimitiveType.GDAY:
               maxLen = 5; // this is the minimum value for a date string using schema syntax
               precision = 2;
               scale = 0;
               break;
           case PrimitiveType.GMONTH:
               maxLen = 6; // this is the minimum value for a date string using schema syntax
               precision = 2;
               scale = 0;
               break;
       }
       
       // check capacity fit
       try
       {
           switch (targetJDBCType)
           {
               case Types.VARCHAR:
               case Types.VARBINARY:
                   
                   if (maxLen < 0)
                   {
                       if (messageHandler != null)
                           messageHandler.warning(new XMLDBCException("XML data may overflow in " + cmeta));
                   }
                   else if (maxLen > cmeta.getColumnSize())
                       throw new SAXException("XML data type length is larger than " + cmeta);
                   break;
                   
               case Types.DECIMAL:
                   if (precision < 0)
                   {
                       if (messageHandler != null)
                           messageHandler.warning(new XMLDBCException("XML data may overflow in " + cmeta));
                   }
                   else if (precision > cmeta.getColumnSize())
                       throw new SAXException("XML data type precision is larger than " + cmeta);
                   else if ((scale >= 0) && (scale > cmeta.getDecimalDigits()))
                       throw new SAXException("XML data type scale is larger than " + cmeta);
                   break;
                   
               default:
           }
       }
       catch (XMLDBCException e)
       {
           throw new SAXException("Operation aborted by user application.", e);
       }
    }
    
    /* check that object other than string (converted) won't lead to a data loss
     * or a storage failure.
     */
    private void checkDataLoss(Object o, XMLErrorHandler messageHandler)
    throws XMLDBCException
    {
        switch(javaType)
        {
            case JAVA_BIG_DECIMAL:
                if (o instanceof BigDecimal && messageHandler != null && cmeta.getColumnSize() > 0)    // system generators (Long, Integer) are not controlled
                {
                    BigDecimal bd = (BigDecimal)o;
                    int precision = (int)Math.ceil(Math.log(bd.unscaledValue().doubleValue())/Math.log(10d));
                    if (precision > cmeta.getColumnSize() ||
                        (cmeta.getDecimalDigits() >= 0 && bd.scale() > cmeta.getDecimalDigits()))
                        messageHandler.warning(new RepositoryException(RepositoryException.DATA_LOSS, "Precision loss."));
                }
                break;
            case JAVA_BYTE_ARRAY:
                if (((byte[])o).length > cmeta.getColumnSize())
                    throw new RepositoryException(RepositoryException.DATA_LOSS, "Value is too large for column.");
                break;
            default:
        }
    }
    
    private void storeConvertedToSQL(PreparedStatement pStmt, int i, Object o)
        throws SQLException, XMLDBCException {
        
        // no check is performed on JDBC types because control has been performed
        // before: conversion is authorized and AUTOMATIC
        switch (javaType) {
			case SCHEMA_DURATION:	// Java types that use strings for JDBC
			case SCHEMA_URI:
			case SCHEMA_QNAME:
			case JAVA_STRING:
			case JAVA_LIST:
                try {
                    pStmt.setBytes(i, ((String) o).getBytes("UTF-8"));
                    // The default XML one
                }
                catch (UnsupportedEncodingException e) {
                    throw new RuntimeException("UTF-8 encoding is not supported by the current JRE. Mapping from string to binary is impossible");
                }
                break;

            case SCHEMA_DATETIME :
				DateTime dt = (DateTime)o;
                int n = 0;
                switch (sType.getPrimitive().getType()) {
                    case PrimitiveType.GYEAR :
                        n = dt.getYear();
                        break;
                    case PrimitiveType.GMONTH :
                        n = dt.getMonth() + 1;
                        break;
                    case PrimitiveType.GDAY :
                        n = dt.getDate();
                        break;
                    default :
                        }
                pStmt.setInt(i, n);
                break;

            default :
                throw new RepositoryException(
                    RepositoryException.INTERNAL_ERROR,
                    "This automatic conversion is not currently supported (javaType: "
                    + javaType + ", JDBCtype: " + cmeta.getDataType() + ')');
        }
    }
    
    private int optimize(int dataType, TypeMap typeMap, long maxLength,
    int precision, int scale, Number min, Number max)
    {
        int optimum = dataType;
        
        switch (dataType) // switch according to "base" target JDBC type
        {   // and try to analyse facets(restrictions) to find a 'smaller one'
            case Types.VARCHAR :
                if (maxLength <= typeMap.getSize(Types.VARCHAR))
                    optimum = Types.VARCHAR;
                else
                    optimum = Types.LONGVARCHAR;
                break;
            case Types.DOUBLE : case Types.FLOAT : // 15 digits mantissa
                if (precision <= 7)
                    optimum = Types.REAL;
                break;
            case Types.NUMERIC :  case Types.DECIMAL : // 15 digits of precision and scale at least
                if (scale == 0) // integers
                {
                    int smallest = getSmallestInteger(precision,
                    min == null ? Double.MIN_VALUE : min.doubleValue(),
                    max == null ? Double.MAX_VALUE : max.doubleValue());
                    if (smallest != 0)
                        optimum = smallest;
                }
                // else use scale and precision coming from XML schema type
                break;
            case Types.VARBINARY :
                if (maxLength <= typeMap.getSize(Types.VARBINARY))
                    optimum = Types.VARBINARY;
                else
                    optimum = Types.LONGVARBINARY;
                break;
            default :   // Nothing to do with VARCHAR, TIMESTAMP, REAL
        }
        return optimum;
    }
    

    /** Calculate the most adapted JDBC integer type basing on the precision & max value given.
     * @return The calculated JDBC-SQL type from java.sql.Types
     * @param precision
     * @param min
     * @param max
     */
    private int getSmallestInteger(int precision, double min, double max)
    {
        int ret = 0;
        int prec = 0;
        
        /* downsizing with boundaries */
        prec = (int)Math.ceil(Math.log(Math.max(Math.abs(min), Math.abs(max)))/Math.log(10d));
        
        if (precision > 0)
            prec = Math.min(precision, prec);
        
        /* downsizing with precision */
        if (prec <= 3)
            ret = Types.TINYINT;
        else if (prec <= 5)
            ret = Types.SMALLINT;
        else if (prec <= 10)
            ret = Types.INTEGER;
        else if (prec <= 19)
            ret = Types.BIGINT;
        
        return ret;
    }
    
    public String toString()
    {
        return "[javaType=" + javaType + ",string=" + isJDBCStringIOAllowed + ",targetJDBC=" + cmeta.getDataType() + "]";
    }
}
